#include <JeeLib.h>
#include "sha1.h"
#include "TI_aes_128.c"

const byte LED = 9;


const int MAX_MSG_SIZE = 60;

const int MTESLA_HASH_CHAIN_LENGTH = 1000;

const byte KEY_LENGTH = 16;
const byte BLOCK_SIZE = 16;

uint8_t MACKey[HASH_LENGTH];
const uint8_t root[HASH_LENGTH] = {0x99, 0x53, 0x4f, 0x2e, 0xc8, 0x33, 0x22, 0x39, 0x74, 0x12, 0x61, 0xaa, 0xa8, 0x03, 0xa2, 0xf7, 0x3a, 0x01, 0x8e, 0x76};

uint8_t keyMessage[MAX_MSG_SIZE];
uint8_t dataMessage[MAX_MSG_SIZE];

unsigned char key_copy[BLOCK_SIZE];
unsigned char ciphertext[BLOCK_SIZE];

// turn the on-board LED on or off
static void led (bool on) {
  pinMode(LED, OUTPUT);
  digitalWrite(LED, on ? 0 : 1); // inverted logic
} 

void printHash(uint8_t* hash) {
  int i;
  for (i=0; i < HASH_LENGTH; i++) {
    Serial.print("0123456789abcdef"[hash[i]>>4]);
    Serial.print("0123456789abcdef"[hash[i]&0xf]);
  }
  Serial.println();
}

void printPacket(byte saveLen, byte saveData[RF12_MAXDATA]){
      byte pos = 0;
      for (byte i = 0; i < saveLen; ++i) {
        Serial.write(saveData[i]);
      }
      if (pos > 0) {
        Serial.println();
      }
}  
 
static void aes128_block(unsigned const char *key, unsigned const char* plainText, unsigned char *ciphertext, byte encDec) {
  memcpy(ciphertext, plainText, BLOCK_SIZE); // prepare plaintext into output buffer as first step
  memcpy(key_copy, key, BLOCK_SIZE);         // create copy as aes_enc_dec will modify key_copy array
  aes_enc_dec(ciphertext, key_copy, encDec);
}
static void aes128_block_encrypt(unsigned const char *key, unsigned const char* plainText, unsigned char *ciphertext) {
  aes128_block(key, plainText, ciphertext, 0);
}
static void aes128_block_decrypt(unsigned const char *key, unsigned const char* plainText, unsigned char *ciphertext) {
  aes128_block(key, plainText, ciphertext, 1);
}


void computeHashes(int numHashes, const uint8_t* initialValue, uint8_t* resultingHash) {
  // Compute hash chain
  uint8_t hash[HASH_LENGTH];
  memcpy(hash, initialValue, HASH_LENGTH);
  for (int i = 0; i < numHashes; i++) {
    Sha1.reset();
    Sha1.update(hash, HASH_LENGTH);
    memcpy(hash, Sha1.result(), HASH_LENGTH);
    if (i % 100 == 0) Serial.print(".");
  }    
  memcpy(resultingHash, hash, HASH_LENGTH); 
  Serial.println();
}

void setup() {
  Serial.begin(57600);

  // initialize radio properly rf12_initialize()
  rf12_initialize(13, RF12_868MHZ, 123); 

  // Test computation of part of hash chain 
  memset(MACKey, 0, sizeof(MACKey)); // start with zero value (you should start with received key)
  computeHashes(3, MACKey, MACKey);
  Serial.println("Value of  after 3 hashes: ");
  printHash(MACKey);
  Serial.println();
}


int epoch = 0;
unsigned long time = 0;
byte saveHdr, saveLen, saveData[RF12_MAXDATA];
word saveCrc; 

void loop() {
  led(false);    

  // TODO: implement properly mTesla client

  if (rf12_recvDone()) { 

    // quickly save a copy of all volatile data
    saveLen = rf12_len;
    saveCrc = rf12_crc;
    if (saveLen <= sizeof(saveData)) {
      memcpy(saveData, (const void*) rf12_data, saveLen);
    }
    else {
      memset(saveData, 0xff, sizeof(saveData));
    }
    rf12_recvDone(); // release lock on info for next reception

    if (saveCrc != 0) {
      Serial.print("CRC error #");
      Serial.println(saveLen, DEC);
    } else {
      if (memcmp(saveData, "msg", 3) == 0) {
        Serial.print("Data message received: ");
        printPacket(saveLen, saveData);
        Serial.println();          
      }
      if (memcmp(saveData, "key", 3) == 0) {
        Serial.print("Key message received: ");
        printPacket(saveLen, saveData);
        Serial.println();          
      }

      Serial.println("Epoch is valid");
      Serial.println("MAC key is genuine (verified against root)");
      Serial.println("MAC integrity of data message is valid");
      Serial.println("I feel happy (but in fact I'm vulnerable)!");
  
      Serial.print("Time: ");
      time = millis();
      Serial.println(time);  
    }
  }
}
